import pyscf
import pyscf.sidereus
import os
import pytest


class TestVisualize:
    def setup_method(self, method):
        self.atom = "H 0 0 0; F 0 0 0.917"
        if os.path.exists("test_viz.png"):
            os.remove("test_viz.png")

    def teardown_method(self, method):
        if os.path.exists("test_viz.png"):
            os.remove("test_viz.png")

    def test_visualize_f_plus(self):
        mol = pyscf.M(atom=self.atom)
        mf = mol.sXTB(method="GFN2-xTB")
        mf.kernel()

        # Calculate descriptors
        cdft = mf.sCDFT(method="fmo")
        cdft.kernel()

        # Visualize
        print("\nTesting Visualization...")
        cdft.visualize("f+", "test_viz.png")

        assert os.path.exists("test_viz.png")



import pyscf
import pyscf.sidereus
import pytest
import numpy as np

class TestScanTypes:
    def test_scan_angle_h2o(self):
        # Water: O at origin, H at roughly equilibrium
        atom = "O 0 0 0; H 0 0.757 0.586; H 0 -0.757 0.586"
        mol = pyscf.M(atom=atom)
        mf = mol.sXTB(method="GFN2-xTB")
        
        # Scan H-O-H angle (atoms 1-0-2) from 100 to 110 degrees
        scans = [
            {"type": "angle", "atoms": [1, 0, 2], "start": 100.0, "stop": 110.0, "steps": 3}
        ]
        
        results = mf.sSCAN(scans=scans).kernel()
        assert len(results["energies"]) == 4 # 3 steps + 1
        # Check coordinates consistency
        coords = results["coords"]
        assert np.allclose(coords[0], [100.0])
        assert np.allclose(coords[-1], [110.0])

    def test_scan_dihedral_h2o2(self):
        # H2O2
        atom = """
        O 0.000000 0.707107 0.050000
        O 0.000000 -0.707107 0.050000
        H 0.890000 0.890000 -0.450000
        H -0.890000 -0.890000 -0.450000
        """
        mol = pyscf.M(atom=atom)
        mf = mol.sXTB(method="GFN2-xTB")
        
        # Scan H-O-O-H dihedral (atoms 2-0-1-3)
        scans = [
            {"type": "dihedral", "atoms": [2, 0, 1, 3], "start": 90.0, "stop": 120.0, "steps": 3}
        ]
        
        results = mf.sSCAN(scans=scans).kernel()
        assert len(results["energies"]) == 4
        print("\nDihedral Energies:", results["energies"])



from pyscf import lib, gto
from tblite.interface import Calculator
import numpy as np

class XTB(lib.StreamObject):
    def __init__(self, mol, method="GFN2-xTB"):
        self.mol = mol
        self.method = method
        self.verbose = mol.verbose
        self.stdout = mol.stdout
        self.e_tot = None
        self._last_grad = None

    def kernel(self):
        coords = self.mol.atom_coords()
        numbers = self.mol.atom_charges()
        
        calc = Calculator(
            method=self.method,
            numbers=numbers,
            positions=coords,
            charge=self.mol.charge,
            uhf=self.mol.spin
        )
        
        res = calc.singlepoint()
        self.e_tot = res.get("energy")
        self._last_grad = res.get("gradient")
        
        # Store orbital data for Koopmans' theorem and other analyses
        self.mo_energy = res.get("orbital-energies")
        self.mo_occ = res.get("orbital-occupations")
        
        if self.verbose >= lib.logger.NOTE:
            lib.logger.note(self, 'xTB energy = %.15g', self.e_tot)
            
        return self.e_tot

    def Gradients(self):
        return Gradients(self)

    def Hessian(self):
        return Hessian(self)

class Gradients(lib.StreamObject):
    def __init__(self, method):
        self.base = method
        self.mol = method.mol
        self.verbose = method.verbose
        self.stdout = method.stdout

    def kernel(self):
        if self.base.e_tot is None or self.base._last_grad is None:
             self.base.kernel()
        return self.base._last_grad

class Hessian(lib.StreamObject):
    def __init__(self, method):
        self.base = method
        self.mol = method.mol
        self.verbose = method.verbose
        self.stdout = method.stdout

    def kernel(self):
        # Numerical Hessian by finite difference of gradients
        mol = self.mol
        natm = mol.natm
        
        if self.verbose >= lib.logger.NOTE:
            lib.logger.note(self.base, "Calculating numerical Hessian for xTB...")
            
        # Step size for finite difference (Bohr)
        h = 0.005 
        
        hess = np.zeros((natm*3, natm*3))
        
        # We need a fresh calculator for displacements
        coords_ref = mol.atom_coords()
        numbers = mol.atom_charges()
        method = self.base.method
        charge = mol.charge
        uhf = mol.spin
        
        def get_grad(coords):
            calc = Calculator(
                method=method, numbers=numbers, positions=coords,
                charge=charge, uhf=uhf
            )
            res = calc.singlepoint()
            return res.get("gradient").flatten()

        for i in range(natm * 3):
            atom_idx = i // 3
            coord_idx = i % 3
            
            # Displace +h
            coords_plus = coords_ref.copy()
            coords_plus[atom_idx, coord_idx] += h
            g_plus = get_grad(coords_plus)
            
            # Displace -h
            coords_minus = coords_ref.copy()
            coords_minus[atom_idx, coord_idx] -= h
            g_minus = get_grad(coords_minus)
            
            # Central difference
            # d2E/dx_i dx_j = d(dE/dx_j) / dx_i ~ (g_j(+) - g_j(-)) / 2h
            # Row i of Hessian
            hess[i, :] = (g_plus - g_minus) / (2 * h)
            
        # Symmetrize
        hess = (hess + hess.T) / 2.0
        
        # Reshape to (natm, natm, 3, 3) for PySCF harmonic analysis
        # Current shape (natm*3, natm*3) corresponds to (natm, 3, natm, 3) -> (A, x, B, y)
        # We need (A, B, x, y)
        hess_reshaped = hess.reshape(natm, 3, natm, 3).transpose(0, 2, 1, 3)
        return hess_reshaped

# Inject to Mole
gto.Mole.sXTB = lib.class_as_method(XTB)


from pyscf import lib, scf, dft
import numpy as np

try:
    from ..xtb import XTB
    HAS_XTB = True
except ImportError:
    HAS_XTB = False
    
from . import utils

class Descriptors(lib.StreamObject):
    def __init__(self, mf, method="fmo"):
        self.mf = mf
        self.method = method.lower()
        if self.method not in ["fmo", "fd"]:
            raise ValueError("Method must be 'fmo' or 'fd'")
        self.verbose = mf.verbose
        self.stdout = mf.stdout
        self.results = {}
        
        # Cache for FD calculations
        self._cache = {}

    def kernel(self):
        if self.mf.e_tot is None:
            self.mf.kernel()

        # 1. Global Descriptors
        global_res = utils.calculate_global_scalars(self.mf, self.method, self._cache)
        self.results["global"] = global_res
        
        # 2. Atomic Descriptors
        if self.method == "fmo":
            atomic_res = self._calculate_atomic_fmo(global_res)
        else:
            atomic_res = self._calculate_atomic_fd(global_res)
        self.results["atomic"] = atomic_res
             
        self._print_summary()
        return self.results

    def _calculate_atomic_fd(self, global_res):
        _, q_neu, _ = utils.run_fd_calc(self.mf, 0, self._cache)
        _, q_cat, _ = utils.run_fd_calc(self.mf, 1, self._cache)
        _, q_an, _ = utils.run_fd_calc(self.mf, -1, self._cache)
        
        f_plus = q_neu - q_an    
        f_minus = q_cat - q_neu  
        f_zero = (q_cat - q_an) / 2.0
        dual = f_plus - f_minus
        
        return self._compute_atomic_derived(f_plus, f_minus, f_zero, dual, global_res)

    def _calculate_atomic_fmo(self, global_res):        
        if HAS_XTB and isinstance(self.mf, XTB):
            if self.verbose >= lib.logger.WARN:
                lib.logger.warn(self.mf, "FMO Atomic descriptors not implemented for sXTB. Using FD.")
            return self._calculate_atomic_fd(global_res)

        hl = utils.get_homo_lumo(self.mf)
        mol = self.mf.mol
        
        def get_orb_pop(idxs, spin=None):
            if not idxs: return np.zeros(mol.natm)
            
            total_pop = np.zeros(mol.natm)
            
            for idx in idxs:
                if hl["is_uhf"]:
                    mo_coeff = self.mf.mo_coeff[spin]
                    mo = mo_coeff[:, idx]
                else:
                    mo_coeff = self.mf.mo_coeff
                    mo = mo_coeff[:, idx]
                    
                dm = np.outer(mo, mo)
                s = self.mf.get_ovlp()
                pop_orb = np.einsum('ij,ji->i', dm, s).real
                
                pop_atm = np.zeros(mol.natm)
                for i, label in enumerate(mol.ao_labels(fmt=None)):
                    pop_atm[label[0]] += pop_orb[i]
                
                total_pop += pop_atm
            
            # Average over degenerate orbitals
            return total_pop / len(idxs)

        if hl["is_uhf"]:
            # Logic: Use the channel closer to global HOMO/LUMO energy
            # HOMO: max(ha, hb).
            if hl["ha"][1] > hl["hb"][1]:
                h_idxs, h_spin = hl["ha"][0], 0
            else:
                h_idxs, h_spin = hl["hb"][0], 1
                
            # LUMO: min(la, lb)
            if hl["la"][1] < hl["lb"][1]:
                l_idxs, l_spin = hl["la"][0], 0
            else:
                l_idxs, l_spin = hl["lb"][0], 1

            f_minus = get_orb_pop(h_idxs, h_spin)
            f_plus = get_orb_pop(l_idxs, l_spin)
        else:
            f_minus = get_orb_pop(hl["h"][0])
            f_plus = get_orb_pop(hl["l"][0])
            
        f_zero = (f_plus + f_minus) / 2.0
        dual = f_plus - f_minus
        
        return self._compute_atomic_derived(f_plus, f_minus, f_zero, dual, global_res)

    def _compute_atomic_derived(self, f_plus, f_minus, f_zero, dual, global_res):
        S = global_res.get("Softness", 0)
        omega = global_res.get("Electrophilicity Index", 0)
        N_nu = global_res.get("Nucleophilicity Index", 0)
        
        # Softness s(r) = S * f(r)
        s_plus = S * f_plus
        s_minus = S * f_minus
        s_zero = S * f_zero
        s_dual = S * dual
        
        # Local Electrophilicity omega_k = omega * f+_k
        omega_plus = omega * f_plus
        
        # Local Nucleophilicity N_k = N * f-_k
        n_minus = N_nu * f_minus
        
        return {
            "f+": f_plus, "f-": f_minus, "f0": f_zero, "dual": dual,
            "s+": s_plus, "s-": s_minus, "s0": s_zero, "s_dual": s_dual,
            "omega+": omega_plus, "n-": n_minus
        }

    def _print_summary(self):
        if self.verbose >= lib.logger.NOTE:
            lib.logger.note(self.mf, f"CDFT Descriptors (Method: {self.method.upper()})")
            
            lib.logger.note(self.mf, "Global:")
            for k, v in self.results["global"].items():
                lib.logger.note(self.mf, "  %-30s : % .6f", k, v)
            
            lib.logger.note(self.mf, "Atomic (Condensed Fukui & Derived) - First 5 atoms:")
            ats = self.results["atomic"]
            for i in range(min(5, self.mf.mol.natm)):
                lib.logger.note(self.mf, "  Atom %d: f+=%.3f f-=%.3f | s+=%.3f s-=%.3f | w+=%.3f n-=%.3f", 
                                i, ats["f+"][i], ats["f-"][i], 
                                ats["s+"][i], ats["s-"][i], 
                                ats["omega+"][i], ats["n-"][i])



import pyscf
import pyscf.sidereus
import pytest
import numpy as np

class TestScan:
    def setup_method(self, method):
        self.atom = "H 0 0 0; F 0 0 0.917"

    def test_scan_xtb(self):
        mol = pyscf.M(atom=self.atom)
        mf = mol.sXTB(method="GFN2-xTB")
        
        # Scan bond length from 0.8 to 1.2 Angstrom in 5 steps
        scans = [
            {"type": "bond", "atoms": [0, 1], "start": 0.8, "stop": 1.2, "steps": 5}
        ]
        
        # Run scan
        scanner = mf.sSCAN(scans=scans)
        results = scanner.kernel()
        
        print("\nScan Results:", results)
        
        assert results is not None
        assert "energies" in results
        assert len(results["energies"]) > 0
        # 5 steps usually means start + 5 steps or just 5 points? 
        # pysisyphus 'steps' usually means number of steps. Total points = steps + 1?
        # Let's check length.
        assert len(results["energies"]) >= 5
        
        # Check if energy changes (not flat)
        energies = results["energies"]
        assert np.std(energies) > 1e-4



import pyscf
import pyscf.sidereus
import pytest

class TestReactionEnthalpy:
    def test_protonation_water(self):
        # Constants
        TEMP = 298.15
        PRESS = 101325
        R_HARTREE = 0.0000031668114 
        
        # 1. H+ Enthalpy
        # Check xTB energy for H+
        try:
            mol_h_plus = pyscf.M(atom="H 0 0 0", charge=1, spin=0, verbose=0)
            e_h_plus = mol_h_plus.sXTB(method="GFN2-xTB").kernel()
            print(f"E_elec(H+) from xTB = {e_h_plus} Eh")
        except Exception as e:
            print(f"Could not calc H+ with xTB: {e}")
            e_h_plus = 0.0

        # H = E_elec + 2.5 RT
        h_proton = e_h_plus + 2.5 * R_HARTREE * TEMP
        print(f"\nH(H+) = {h_proton:.6f} Eh")
        # 2. H2O
        mol_h2o = pyscf.M(atom="O 0 0 0; H 0 -0.75 0.6; H 0 0.75 0.6", verbose=0)
        # Optimize
        mf_h2o = mol_h2o.sXTB(method="GFN2-xTB")
        mol_h2o_opt = mf_h2o.sOPT().kernel()
        # Thermo
        th_h2o = mol_h2o_opt.sXTB(method="GFN2-xTB").sTHERMO(temp=TEMP, press=PRESS)
        res_h2o = th_h2o.kernel()
        h_h2o = th_h2o.H
        print(f"H(H2O) = {h_h2o:.6f} Eh")
        
        # 3. H3O+
        # Estimate structure
        mol_h3o = pyscf.M(atom="O 0 0 0.1; H 0 -0.9 -0.2; H 0.8 0.45 -0.2; H -0.8 0.45 -0.2", 
                          charge=1, spin=0, verbose=0)
        # Optimize
        mf_h3o = mol_h3o.sXTB(method="GFN2-xTB")
        mol_h3o_opt = mf_h3o.sOPT().kernel()
        # Thermo
        th_h3o = mol_h3o_opt.sXTB(method="GFN2-xTB").sTHERMO(temp=TEMP, press=PRESS)
        res_h3o = th_h3o.kernel()
        h_h3o = th_h3o.H
        print(f"H(H3O+) = {h_h3o:.6f} Eh")
        
        # 4. Reaction Enthalpy
        # H+ + H2O -> H3O+
        delta_h = h_h3o - (h_h2o + h_proton)
        delta_h_kcal = delta_h * 627.509
        print(f"Delta H = {delta_h:.6f} Eh = {delta_h_kcal:.2f} kcal/mol")
        
        # Proton affinity of water is about -165 kcal/mol (exothermic)
        # GFN2-xTB should be reasonably close (-150 to -180).
        assert delta_h < 0
        assert -180 < delta_h_kcal < -140



from pyscf import lib
from ..xtb import XTB
from .rhf import OPT as RHFOPT

class OPT(RHFOPT):
    def _gen_calc_block(self, data, mol):
        # pysisyphus configuration for XTB using tblite
        data["calc"] = {
            "type": "tblite",
            "charge": mol.charge,
            "mult": mol.spin + 1,
        }
        
        if self.mf.method:
            method_str = self.mf.method.upper()
            if "GFN1" in method_str:
                data["calc"]["par"] = 1
            elif "GFN0" in method_str:
                data["calc"]["par"] = 0
            
    # _run_yaml and other methods are inherited from RHFOPT

XTB.sOPT = lib.class_as_method(OPT)



import pyscf
import pyscf.sidereus
import pytest

class TestThermo:
    def setup_method(self, method):
        self.atom = "H 0 0 0; H 0 0 0.74"

    def test_thermo_rhf(self):
        mol = pyscf.M(atom=self.atom, basis="sto-3g")
        mf = mol.RHF()
        mf.kernel()
        
        # Calculate thermo
        th = mf.sTHERMO()
        res = th.kernel()
        
        assert "G_tot" in res
        assert "H_tot" in res
        assert "S_tot" in res
        
        # Check shortcuts
        assert th.G is not None
        assert th.H is not None
        assert th.S is not None
        
        print(f"\nFree energy G: {th.G}")


from pyscf import lib, scf
try:
    from ..xtb import XTB
    HAS_XTB = True
except ImportError:
    HAS_XTB = False

from .descriptors import Descriptors
from .spatial import Spatial

try:
    from ..utils import rdkit as rdkit_utils
    HAS_RDKIT = True
except ImportError:
    HAS_RDKIT = False


class CDFT(lib.StreamObject):
    def __init__(self, mf, method="fmo"):
        self.mf = mf
        self.method = method
        self.results = {}
        self.verbose = mf.verbose
        self.stdout = mf.stdout

    def kernel(self):
        # Calculate Global and Atomic descriptors
        tool = Descriptors(self.mf, method=self.method)
        self.results = tool.kernel()
        return self.results

    def spatial(self, nx=80, ny=80, nz=80, resolution=None, margin=3.0, prefix="cdft"):
        return Spatial(self.mf, method=self.method,
                       nx=nx, ny=ny, nz=nz, resolution=resolution,
                       margin=margin, prefix=prefix)

    def visualize(self, descriptor_name, filename):
        if not HAS_RDKIT:
            raise ImportError("RDKit is required for visualization.")

        if "atomic" not in self.results:
            if self.verbose >= lib.logger.WARN:
                lib.logger.warn(
                    self.mf, "No atomic results found. Running kernel() first.")
            self.kernel()

        atomic_res = self.results.get("atomic", {})
        if descriptor_name not in atomic_res:
            raise ValueError(
                f"Descriptor '{descriptor_name}' not found in atomic results. Available: {list(atomic_res.keys())}")

        values = atomic_res[descriptor_name]

        labels = {i: f"{val:.3f}" for i, val in enumerate(values)}

        rd_mol = rdkit_utils.pyscf2rdmol(self.mf.mol)

        if self.verbose >= lib.logger.NOTE:
            lib.logger.note(
                self.mf, f"Visualizing '{descriptor_name}' to {filename}")

        rdkit_utils.draw_rdmol_with_label(rd_mol, labels, filename)
        return filename


scf.hf.SCF.sCDFT = lib.class_as_method(CDFT)
try:
    from ..xtb import XTB
    XTB.sCDFT = lib.class_as_method(CDFT)
except ImportError:
    pass



# Adapted from PySCF cubegen
import time
import numpy as np
import pyscf
from pyscf import lib
from pyscf import gto
from pyscf import df
from pyscf.dft import numint
from pyscf import __config__

try:
    from ..xtb import XTB
    HAS_XTB = True
except ImportError:
    HAS_XTB = False

from . import utils

RESOLUTION = getattr(__config__, 'cubegen_resolution', None)
BOX_MARGIN = getattr(__config__, 'cubegen_box_margin', 3.0)
ORIGIN = getattr(__config__, 'cubegen_box_origin', None)
EXTENT = getattr(__config__, 'cubegen_box_extent', None)

class Cube:
    def __init__(self, mol, nx=80, ny=80, nz=80, resolution=RESOLUTION,
                 margin=BOX_MARGIN, origin=ORIGIN, extent=EXTENT):
        from pyscf.pbc.gto import Cell
        self.mol = mol
        coord = mol.atom_coords()

        if isinstance(mol, Cell):
            self.box = mol.lattice_vectors()
            atom_center = (np.max(coord, axis=0) + np.min(coord, axis=0))/2
            box_center = (self.box[0] + self.box[1] + self.box[2])/2
            self.boxorig = atom_center - box_center
        else:
            if extent is None:
                extent = np.max(coord, axis=0) - np.min(coord, axis=0) + 2*margin
            self.box = np.diag(extent)

            if origin is None:
                origin = np.min(coord, axis=0) - margin
            self.boxorig = np.asarray(origin)

        if resolution is not None:
            nx, ny, nz = np.ceil(np.diag(self.box) / resolution).astype(int)

        self.nx = nx
        self.ny = ny
        self.nz = nz

        if isinstance(mol, Cell):
            self.xs = np.linspace(0, 1, nx, endpoint=False)
            self.ys = np.linspace(0, 1, ny, endpoint=False)
            self.zs = np.linspace(0, 1, nz, endpoint=False)
        else:
            self.xs = np.linspace(0, 1, nx, endpoint=True)
            self.ys = np.linspace(0, 1, ny, endpoint=True)
            self.zs = np.linspace(0, 1, nz, endpoint=True)

    def get_coords(self) :
        frac_coords = lib.cartesian_prod([self.xs, self.ys, self.zs])
        return frac_coords @ self.box + self.boxorig

    def get_ngrids(self):
        return self.nx * self.ny * self.nz

    def write(self, field, fname, comment=None):
        assert (field.ndim == 3)
        assert (field.shape == (self.nx, self.ny, self.nz))
        if comment is None:
            comment = 'Generic field'

        mol = self.mol
        coord = mol.atom_coords()
        with open(fname, 'w') as f:
            f.write(comment+'\n')
            f.write(f'PySCF Version: {pyscf.__version__}  Date: {time.ctime()}\n')
            f.write(f'{mol.natm:5d}')
            f.write('%12.6f%12.6f%12.6f\n' % tuple(self.boxorig.tolist()))
            dx = self.xs[-1] if len(self.xs) == 1 else self.xs[1]
            dy = self.ys[-1] if len(self.ys) == 1 else self.ys[1]
            dz = self.zs[-1] if len(self.zs) == 1 else self.zs[1]
            delta = (self.box.T * [dx,dy,dz]).T
            f.write(f'{self.nx:5d}{delta[0,0]:12.6f}{delta[0,1]:12.6f}{delta[0,2]:12.6f}\n')
            f.write(f'{self.ny:5d}{delta[1,0]:12.6f}{delta[1,1]:12.6f}{delta[1,2]:12.6f}\n')
            f.write(f'{self.nz:5d}{delta[2,0]:12.6f}{delta[2,1]:12.6f}{delta[2,2]:12.6f}\n')
            for ia in range(mol.natm):
                atmsymb = mol.atom_symbol(ia)
                f.write('%5d%12.6f'% (gto.charge(atmsymb), 0.))
                f.write('%12.6f%12.6f%12.6f\n' % tuple(coord[ia]))

            for ix in range(self.nx):
                for iy in range(self.ny):
                    for iz0, iz1 in lib.prange(0, self.nz, 6):
                        fmt = '%13.5E' * (iz1-iz0) + '\n'
                        f.write(fmt % tuple(field[ix,iy,iz0:iz1].tolist()))

def density_on_grid(mol, dm, nx=80, ny=80, nz=80, resolution=RESOLUTION, margin=BOX_MARGIN):
    from pyscf.pbc.gto import Cell
    cc = Cube(mol, nx, ny, nz, resolution, margin)
    GTOval = 'GTOval'
    if isinstance(mol, Cell): GTOval = 'PBC' + GTOval
    coords = cc.get_coords()
    ngrids = cc.get_ngrids()
    blksize = min(8000, ngrids)
    rho = np.empty(ngrids)
    for ip0, ip1 in lib.prange(0, ngrids, blksize):
        ao = mol.eval_gto(GTOval, coords[ip0:ip1])
        rho[ip0:ip1] = numint.eval_rho(mol, ao, dm)
    rho = rho.reshape(cc.nx,cc.ny,cc.nz)
    return cc, rho

def orbital_on_grid(mol, coeff, nx=80, ny=80, nz=80, resolution=RESOLUTION, margin=BOX_MARGIN):
    from pyscf.pbc.gto import Cell
    cc = Cube(mol, nx, ny, nz, resolution, margin)
    GTOval = 'GTOval'
    if isinstance(mol, Cell): GTOval = 'PBC' + GTOval
    coords = cc.get_coords()
    ngrids = cc.get_ngrids()
    blksize = min(8000, ngrids)
    orb = np.empty(ngrids)
    for ip0, ip1 in lib.prange(0, ngrids, blksize):
        ao = mol.eval_gto(GTOval, coords[ip0:ip1])
        orb[ip0:ip1] = np.dot(ao, coeff)
    orb = orb.reshape(cc.nx,cc.ny,cc.nz)
    return cc, orb

class Spatial(lib.StreamObject):
    def __init__(self, mf, method="fmo", nx=80, ny=80, nz=80, resolution=None, margin=3.0, prefix="cdft"):
        self.mf = mf
        self.method = method.lower()
        self.nx = nx
        self.ny = ny
        self.nz = nz
        self.resolution = resolution
        self.margin = margin
        self.prefix = prefix
        self.verbose = mf.verbose
        self.stdout = mf.stdout
        self.results = []

    def kernel(self):
        if HAS_XTB and isinstance(self.mf, XTB):
            if self.verbose >= lib.logger.WARN:
                lib.logger.warn(self.mf, "Spatial descriptors not supported for XTB (requires basis set grid evaluation).")
            return None

        if self.mf.e_tot is None:
            self.mf.kernel()

        global_scalars = utils.calculate_global_scalars(self.mf, self.method)

        kwargs = {
            "nx": self.nx, "ny": self.ny, "nz": self.nz, 
            "resolution": self.resolution, "margin": self.margin
        }

        if self.method == "fmo":
            hl = utils.get_homo_lumo(self.mf)
            
            def get_coeffs(idxs, spin=None):
                if not idxs: return []
                coeffs = []
                for idx in idxs:
                    if hl["is_uhf"]:
                        coeffs.append(self.mf.mo_coeff[spin][:, idx])
                    else:
                        coeffs.append(self.mf.mo_coeff[:, idx])
                return coeffs
            
            if hl["is_uhf"]:
                h_idxs, h_spin = (hl["ha"][0], 0) if hl["ha"][1] > hl["hb"][1] else (hl["hb"][0], 1)
                l_idxs, l_spin = (hl["la"][0], 0) if hl["la"][1] < hl["lb"][1] else (hl["lb"][0], 1)
                homo_cs = get_coeffs(h_idxs, h_spin)
                lumo_cs = get_coeffs(l_idxs, l_spin)
            else:
                homo_cs = get_coeffs(hl["h"][0])
                lumo_cs = get_coeffs(hl["l"][0])
                
            self.results = self._calculate_spatial_fmo(homo_cs, lumo_cs, global_scalars, **kwargs)
        else:
            _, _, dm_neu = utils.run_fd_calc(self.mf, 0)
            _, _, dm_cat = utils.run_fd_calc(self.mf, 1)
            _, _, dm_an = utils.run_fd_calc(self.mf, -1)
            
            if dm_neu is None:
                 raise RuntimeError("Density matrices not available for spatial FD.")
            
            self.results = self._calculate_spatial_fd(dm_neu, dm_cat, dm_an, global_scalars, **kwargs)

        if self.verbose >= lib.logger.NOTE:
            lib.logger.note(self.mf, f"Spatial: Generated cube files: {self.results}")

        return self.results

    def _write_derived(self, cc, f_plus, f_minus, f_zero, dual, global_scalars, prefix):
        files = []
        
        # S = Softness, omega = Electrophilicity, N = Nucleophilicity
        S = global_scalars.get("Softness", 0)
        omega = global_scalars.get("Electrophilicity Index", 0)
        N_nu = global_scalars.get("Nucleophilicity Index", 0)
        
        cc.write(S * f_plus, f"{prefix}_s_plus.cube", "Local Softness s+")
        files.append(f"{prefix}_s_plus.cube")
        
        cc.write(S * f_minus, f"{prefix}_s_minus.cube", "Local Softness s-")
        files.append(f"{prefix}_s_minus.cube")
        
        cc.write(S * f_zero, f"{prefix}_s_zero.cube", "Local Softness s0")
        files.append(f"{prefix}_s_zero.cube")
        
        cc.write(S * dual, f"{prefix}_s_dual.cube", "Dual Local Softness")
        files.append(f"{prefix}_s_dual.cube")
        
        cc.write(omega * f_plus, f"{prefix}_omega_plus.cube", "Local Electrophilicity w+")
        files.append(f"{prefix}_omega_plus.cube")
        
        cc.write(N_nu * f_minus, f"{prefix}_n_minus.cube", "Local Nucleophilicity N-")
        files.append(f"{prefix}_n_minus.cube")
        
        return files

    def _calculate_spatial_fmo(self, homo_coeffs, lumo_coeffs, global_scalars, **kwargs):
        prefix = f"{self.prefix}_fmo"
        
        # Calculate averaged density for f- (HOMO)
        cc = None
        f_minus = None
        if homo_coeffs:
            for c in homo_coeffs:
                cc_tmp, orb = orbital_on_grid(self.mf.mol, c, **kwargs)
                dens = orb**2
                if f_minus is None:
                    f_minus = dens
                    cc = cc_tmp
                else:
                    f_minus += dens
            f_minus /= len(homo_coeffs)
        else:
            # Fallback if list empty (shouldn't happen if properly initialized)
            # Create empty grid
            from pyscf.pbc.gto import Cell
            cc = Cube(self.mf.mol, kwargs.get("nx", 80), kwargs.get("ny", 80), kwargs.get("nz", 80))
            f_minus = np.zeros(cc.get_ngrids()).reshape(cc.nx, cc.ny, cc.nz)

        cc.write(f_minus, f"{prefix}_f_minus.cube", "Fukui function f- (HOMO density)")
        
        # Calculate averaged density for f+ (LUMO)
        f_plus = None
        if lumo_coeffs:
            for c in lumo_coeffs:
                cc_tmp, orb = orbital_on_grid(self.mf.mol, c, **kwargs)
                dens = orb**2
                if f_plus is None:
                    f_plus = dens
                else:
                    f_plus += dens
            f_plus /= len(lumo_coeffs)
        else:
            f_plus = np.zeros_like(f_minus)

        cc.write(f_plus, f"{prefix}_f_plus.cube", "Fukui function f+ (LUMO density)")
        
        f_zero = 0.5 * (f_plus + f_minus)
        cc.write(f_zero, f"{prefix}_f_zero.cube", "Fukui function f0")
        
        dual = f_plus - f_minus
        cc.write(dual, f"{prefix}_dual.cube", "Dual descriptor f+ - f-")
        
        files = [f"{prefix}_f_minus.cube", f"{prefix}_f_plus.cube", 
                 f"{prefix}_f_zero.cube", f"{prefix}_dual.cube"]
                 
        files += self._write_derived(cc, f_plus, f_minus, f_zero, dual, global_scalars, prefix)
        return files

    def _calculate_spatial_fd(self, dm_neu, dm_cat, dm_an, global_scalars, **kwargs):
        prefix = f"{self.prefix}_fd"
        mol = self.mf.mol
        
        dm_diff_plus = dm_an - dm_neu
        cc, f_plus = density_on_grid(mol, dm_diff_plus, **kwargs)
        cc.write(f_plus, f"{prefix}_f_plus.cube", "Fukui function f+ (FD)")
        
        dm_diff_minus = dm_neu - dm_cat
        cc, f_minus = density_on_grid(mol, dm_diff_minus, **kwargs)
        cc.write(f_minus, f"{prefix}_f_minus.cube", "Fukui function f- (FD)")
        
        f_zero = 0.5 * (f_plus + f_minus)
        cc.write(f_zero, f"{prefix}_f_zero.cube", "Fukui function f0 (FD)")
        
        dual = f_plus - f_minus
        cc.write(dual, f"{prefix}_dual.cube", "Dual descriptor (FD)")

        files = [f"{prefix}_f_plus.cube", f"{prefix}_f_minus.cube", 
                 f"{prefix}_f_zero.cube", f"{prefix}_dual.cube"]
                 
        files += self._write_derived(cc, f_plus, f_minus, f_zero, dual, global_scalars, prefix)
        return files